<?php

namespace OCA\PahekoFSpace\Controller;

use DateInterval;
use DateTime;
use OCA\PahekoFSpace\AppInfo\Application;
use OCA\PahekoFSpace\Service\SyncUsers;
use OCP\AppFramework\Controller;
use OCP\AppFramework\Http\Attribute\NoAdminRequired;
use OCP\AppFramework\Http\Attribute\NoCSRFRequired;
use OCP\AppFramework\Http\ContentSecurityPolicy;
use OCP\AppFramework\Http\TemplateResponse;
use OCP\IAppConfig;
use OCP\IGroupManager;
use OCP\IRequest;
use OCP\IUserSession;
use ParagonIE\Paseto\Builder;
use ParagonIE\Paseto\Keys\SymmetricKey;
use ParagonIE\Paseto\Protocol\Version4;
use ParagonIE\Paseto\Purpose;

class PageController extends Controller {
	public function __construct(
		$appName,
		IRequest $request,
		private readonly IUserSession $userSession,
		private readonly IGroupManager $groupManager,
		private readonly IAppConfig $config,
		private readonly SyncUsers $syncUsers,
	) {
		parent::__construct($appName, $request);
	}

	#[NoAdminRequired]
	#[NoCSRFRequired]
	public function index(): TemplateResponse {
		$user = $this->userSession->getUser();
		if ($user === null) {
			throw new \Exception('User is not logged in.'); // Should not happen, since method is guarded by middleware
		}

		$sharedKey = $this->config->getValueString(Application::APP_ID, 'api_shared_key');
		if (!$sharedKey) {
			throw new \Exception("Couldn't find shared Paheko key");
		}
		if (!$this->config->getValueBool(Application::APP_ID, 'initial_user_import')) {
			$this->syncUsers->initialSync();
		}
		$pahekoUrl = $this->config->getValueString(Application::APP_ID, 'paheko_url');

		$urlData = parse_url($pahekoUrl);
		$domain = $urlData['host'] ?? null;
		$port = $urlData['port'] ?? null;

		if ($domain === null || $domain === '') {
			throw new \Exception("Couldn't find domain from set Paheko URL");
		}



		if (is_integer($port) && !in_array($port, [80, 443], true)) {
			$domain .= ':' . $port;
		}

		$token = (new Builder())
			->setKey(SymmetricKey::fromEncodedString($sharedKey))
			->setVersion(new Version4)
			->setPurpose(Purpose::local())
			// Set it to expire in one day
			->setIssuedAt()
			->setNotBefore()
			->setExpiration(
				(new DateTime())->add(new DateInterval('P01D'))
			)
			// Store arbitrary data
			->setClaims([
				'username' => $user->getDisplayName(),
				'email' => $user->getEMailAddress(),
				'isAdmin' => $this->groupManager->isAdmin($user->getUID()),
				'permission_accounting' => $this->groupManager->isInGroup($user->getUID(), 'paheko_compta'),
				'permission_members' => $this->groupManager->isInGroup($user->getUID(), 'paheko_membres'),
			]);

		$nonce = sha1(random_bytes(8));

		$oTemplate = new TemplateResponse('paheko_fspace', 'index', ['token' => $token, 'pahekoURL' => $pahekoUrl, 'nonce' => $nonce]);
		$csp = new ContentSecurityPolicy();
		$csp->addAllowedFrameDomain("'self' fossil.kd2.org paheko.cloud");
		$csp->addAllowedFrameDomain($domain);
		$csp->addAllowedScriptDomain("'nonce-$nonce'");
		$csp->addAllowedFormActionDomain($domain);
		$oTemplate->setContentSecurityPolicy($csp);
		return $oTemplate;
	}
}
