<?php

namespace OCA\PahekoFSpace\Service;

use Exception;
use JsonException;
use OCA\PahekoFSpace\AppInfo\Application;
use OCP\Exceptions\AppConfigTypeConflictException;
use OCP\Http\Client\IClient;
use OCP\Http\Client\IClientService;
use OCP\IAppConfig;
use OCP\IConfig;
use OCP\IGroupManager;
use OCP\IUser;
use OCP\IUserManager;
use OCP\PreConditionNotMetException;
use Psr\Log\LoggerInterface;

class SyncUsers {
	private IClient $client;
	private string $pahekoUrl;
	private string $pahekoCredentials;
	private array $excludedUids;
	private string $adminCategoryName;

	public function __construct(
		private readonly IAppConfig $appConfig,
		private readonly IConfig $config,
		private readonly IUserManager $userManager,
		readonly IClientService $clientService,
		private readonly IGroupManager $groupManager,
		private readonly LoggerInterface $logger,
	) {
		$this->client = $this->clientService->newClient();
		$this->pahekoUrl = $this->appConfig->getValueString(Application::APP_ID, 'paheko_url');
		$this->setCredentials();
		$this->excludedUids = explode(',', $this->appConfig->getValueString(Application::APP_ID, 'excluded_uids', 'framadmin'));
		$this->adminCategoryName = $this->appConfig->getValueString(Application::APP_ID, 'admin_category_name', 'Administrateurs');
	}

	/**
	 * @throws AppConfigTypeConflictException
	 * @throws JsonException
	 */
	public function initialSync(): void {
		$initialSyncDone = $this->appConfig->getValueBool(Application::APP_ID, 'initial_user_import');
		if ($initialSyncDone) {
			return;
		}
		$categories = $this->getCategories();

		$this->userManager->callForAllUsers(function (IUser $user) use ($categories) {
			$this->doCreateUser($user, $categories);
			$this->logger->debug('Created user ' . $user->getUID() . ' on Paheko');
		});
		$this->logger->info('Performed Paheko initial user sync');
		$this->appConfig->setValueBool(Application::APP_ID, 'initial_user_import', true);
	}

	/**
	 * @throws JsonException
	 * @throws PreConditionNotMetException
	 */
	public function createUser(IUser $user): void {
		$categories = $this->getCategories();
		$this->doCreateUser($user, $categories);
		$this->logger->info('Created user ' . $user->getUID() . ' on Paheko');
	}

	/**
	 * @throws Exception
	 */
	public function updateUser(IUser $user): void {
		$uid = $user->getUID();
		$pahekoUserId = $this->config->getUserValue($uid, Application::APP_ID, 'paheko_user_id');
		if (!$pahekoUserId || in_array($uid, $this->excludedUids, true)) {
			return;
		}
		$this->logger->debug('Updating user ' . $uid . ' on Paheko');

		try {
			$this->client->post($this->pahekoUrl . '/api/user/' . $pahekoUserId, [
				'body' => json_encode([
					'nom' => $user->getDisplayName() ?? $user->getUID(),
					'email' => $user->getEMailAddress(),
				]),
				'nextcloud' => [
					'allow_local_address' => true,
				],
				'headers' => [
					'Authorization' => 'Basic ' . $this->pahekoCredentials,
					'Content-Type' => 'application/json'
				]
			]);
			$this->logger->info('Updated user ' . $uid . ' on Paheko');
		} catch (Exception $e) {
			$this->logger->error('Error while updating user ' . $uid . ' on Paheko', ['exception' => $e]);
			throw $e;
		}
	}

	/**
	 * @throws Exception
	 */
	public function deleteUser(IUser $user): void {
		$uid = $user->getUID();
		$pahekoUserId = $this->config->getUserValue($uid, Application::APP_ID, 'paheko_user_id');
		if (!$pahekoUserId || in_array($uid, $this->excludedUids, true)) {
			return;
		}
		$this->logger->debug('Deleting user ' . $uid . ' on Paheko');

		try {
			$this->client->delete($this->pahekoUrl . '/api/user/' . $pahekoUserId, [
				'nextcloud' => [
					'allow_local_address' => true,
				],
				'headers' => [
					'Authorization' => 'Basic ' . $this->pahekoCredentials,
					'Content-Type' => 'application/json'
				]
			]);
			$this->logger->info('Deleted user ' . $uid . ' on Paheko');
		} catch (Exception $e) {
			$this->logger->error('Error while deleting user ' . $uid . ' on Paheko', ['exception' => $e]);
			throw $e;
		}
	}

	/**
	 * @throws JsonException
	 * @throws PreConditionNotMetException
	 * @throws Exception
	 */
	private function doCreateUser(IUser $user, array $categories): void {
		$uid = $user->getUID();
		if (in_array($uid, $this->excludedUids, true)) {
			return;
		}
		try {
			$isAdmin = $this->groupManager->isAdmin($uid);
			$this->logger->debug('Creating user ' . $uid . ($isAdmin ? ' (admin)' : '') . ' on Paheko');

			$response = $this->client->post($this->pahekoUrl . '/api/user/new', [
				'body' => json_encode([
					'nom' => $user->getDisplayName() ?? $user->getUID(),
					'email' => $user->getEMailAddress(),
					'id_category' => $isAdmin ? $categories['admin'] : $categories['active'],
				]),
				'nextcloud' => [
					'allow_local_address' => true,
				],
				'headers' => [
					'Authorization' => 'Basic ' . $this->pahekoCredentials,
					'Content-Type' => 'application/json'
				]
			]);
			/** @var array{id: integer} $data */
			$data = json_decode($response->getBody(), true, 512, JSON_THROW_ON_ERROR);
			$this->config->setUserValue($uid, Application::APP_ID, 'paheko_user_id', (string)$data['id']);
		} catch (Exception $e) {
			if ($e->getCode() !== 409) {
				$this->logger->error('Error while creating user ' . $uid . ' on Paheko', ['exception' => $e]);
				throw $e;
			} else {
				$this->logger->info('Conflict while trying to create user ' . $uid . ' on Paheko', ['exception' => $e]);
			}
		}
	}

	private function getCategories(): array {
		try {
			$response = $this->client->get($this->pahekoUrl . '/api/user/categories', [
				'nextcloud' => [
					'allow_local_address' => true,
				],
				'headers' => [
					'Authorization' => 'Basic ' . $this->pahekoCredentials,
					'Content-Type' => 'application/json'
				]
			]);
			/** @var array{id: integer, name: string}[] $categories */
			$categories = array_values(json_decode($response->getBody(), true, 512, JSON_THROW_ON_ERROR));
			$adminCategory = array_values(array_filter($categories, fn ($category) => $category['name'] === $this->adminCategoryName))[0];
			$activeCategory = array_values(array_filter($categories, fn ($category) => $category['name'] === 'Membres actifs'))[0];
			return ['admin' => $adminCategory['id'], 'active' => $activeCategory['id']];
		} catch (Exception $e) {
			$this->logger->error('Error while fetching Paheko user categories', ['exception' => $e]);
			throw $e;
		}
	}

	/**
	 * @throws AppConfigTypeConflictException
	 */
	private function setCredentials(): void {
		$pahekoAPIUser = $this->appConfig->getValueString(Application::APP_ID, 'paheko_api_user');
		$pahekoAPIPassword = $this->appConfig->getValueString(Application::APP_ID, 'paheko_api_password');
		if (!$pahekoAPIUser || !$pahekoAPIPassword) {
			throw new Exception('Paheko API user and/or password is not configured');
		}
		$this->pahekoCredentials = base64_encode($pahekoAPIUser . ':' . $pahekoAPIPassword);
	}
}
