<?php

/**
 * SPDX-FileCopyrightText: 2024 Framasoft <https://framasoft.org>
 * SPDX-FileContributor: Thomas Citharel <thomas.citharel@framasoft.org>
 *
 * SPDX-License-Identifier: AGPL-3.0-only
 */

declare(strict_types=1);

namespace OCA\PahekoFSpace\Tests\Service;

use OCA\PahekoFSpace\AppInfo\Application;
use OCA\PahekoFSpace\Controller\PageController;
use OCA\PahekoFSpace\Service\SyncUsers;
use OCP\AppFramework\Http\TemplateResponse;
use OCP\IAppConfig;
use OCP\IGroupManager;
use OCP\IRequest;
use OCP\IUser;
use OCP\IUserSession;
use PHPUnit\Framework\Attributes\DataProvider;
use PHPUnit\Framework\MockObject\MockObject;
use Test\TestCase;

class PageControllerTest extends TestCase {

	private IRequest|MockObject $request;
	private IUserSession|MockObject $userSession;
	private IGroupManager|MockObject $groupManager;
	private IAppConfig|MockObject $config;
	private SyncUsers|MockObject $syncUsers;
	private PageController $controller;

	public function setUp(): void {
		parent::setUp();

		$this->request = $this->createMock(IRequest::class);
		$this->userSession = $this->createMock(IUserSession::class);
		$this->groupManager = $this->createMock(IGroupManager::class);
		$this->config = $this->createMock(IAppConfig::class);
		$this->syncUsers = $this->createMock(SyncUsers::class);

		$this->controller = new PageController(Application::APP_ID, $this->request, $this->userSession, $this->groupManager, $this->config, $this->syncUsers);
	}

	public function testIndex() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->exactly(3))->method('getUID')->willReturn('someuser');
		$user->expects($this->once())->method('getDisplayName')->willReturn('Some User');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('some@user.fr');

		$this->userSession->expects($this->once())->method('getUser')->willReturn($user);
		$matcher = $this->exactly(2);
		$this->config->expects($matcher)->method('getValueString')->willReturnCallback(function () use ($matcher) {
			return match ($matcher->numberOfInvocations()) {
				1 => [Application::APP_ID, 'api_shared_key'],
				2 => [Application::APP_ID, 'paheko_url'],
			};
		})->willReturnOnConsecutiveCalls('l-aY2VxvqiWJDQUcpYbSgbdtubzNE9BpJI7nMtuW_Nk', 'https://paheko.domain.frama.space');

		$this->groupManager->expects($this->once())->method('isAdmin')->willReturn(false);

		$response = $this->controller->index();

		$this->assertInstanceOf(TemplateResponse::class, $response);
		$csp = $response->getContentSecurityPolicy()->buildPolicy();
		$this->assertStringContainsString("script-src 'self' 'nonce-", $csp);
		$this->assertStringContainsString("frame-src 'self' fossil.kd2.org paheko.cloud paheko.domain.frama.space", $csp);
		$this->assertStringContainsString("form-action 'self' paheko.domain.frama.space", $csp);
	}

	public function testWithoutBeingLoggedIn() {
		$this->userSession->expects($this->once())->method('getUser')->willReturn(null);

		$this->expectExceptionMessage('User is not logged in.');
		$this->controller->index();
	}

	public function testWithoutKeySet() {
		$user = $this->createMock(IUser::class);
		$this->userSession->expects($this->once())->method('getUser')->willReturn($user);

		$this->config->expects($this->once())->method('getValueString')->with(Application::APP_ID, 'api_shared_key')->willReturn('');
		$this->expectExceptionMessage("Couldn't find shared Paheko key");
		$this->controller->index();
	}

	#[DataProvider('dataTestWithBadPahekoURL')]
	public function testWithBadPahekoURL(string $pahekoURL) {
		$user = $this->createMock(IUser::class);
		$this->userSession->expects($this->once())->method('getUser')->willReturn($user);

		$matcher = $this->exactly(2);
		$this->config->expects($matcher)->method('getValueString')->willReturnCallback(function () use ($matcher) {
			return match ($matcher->numberOfInvocations()) {
				1 => [Application::APP_ID, 'api_shared_key'],
				2 => [Application::APP_ID, 'paheko_url'],
			};
		})->willReturnOnConsecutiveCalls('l-aY2VxvqiWJDQUcpYbSgbdtubzNE9BpJI7nMtuW_Nk', $pahekoURL);
		$this->expectExceptionMessage("Couldn't find domain from set Paheko URL");
		$this->controller->index();
	}

	public static function dataTestWithBadPahekoURL(): array {
		return [
			[''],
			['not an URL']
		];
	}
}
