<?php

/**
 * SPDX-FileCopyrightText: 2024 Framasoft <https://framasoft.org>
 * SPDX-FileContributor: Thomas Citharel <thomas.citharel@framasoft.org>
 *
 * SPDX-License-Identifier: AGPL-3.0-only
 */

declare(strict_types=1);

namespace OCA\PahekoFSpace\Tests\Service;

use OCA\PahekoFSpace\AppInfo\Application;
use OCA\PahekoFSpace\Service\SyncUsers;
use OCP\Http\Client\IClient;
use OCP\Http\Client\IClientService;
use OCP\Http\Client\IResponse;
use OCP\IAppConfig;
use OCP\IConfig;
use OCP\IGroupManager;
use OCP\IUser;
use OCP\IUserManager;
use PHPUnit\Framework\Attributes\DataProvider;
use PHPUnit\Framework\MockObject\MockObject;
use Psr\Log\LoggerInterface;
use Test\TestCase;

class SyncUsersTest extends TestCase {
	private SyncUsers $syncUsers;

	private IAppConfig|MockObject $appConfig;
	private IConfig|MockObject $config;
	private IUserManager|MockObject $userManager;
	private IClientService|MockObject $clientService;
	private IClient|MockObject $client;
	private IGroupManager|MockObject $groupManager;
	private LoggerInterface|MockObject $logger;

	public function setUp(): void {
		parent::setUp();

		$this->appConfig = $this->createMock(IAppConfig::class);
		$this->config = $this->createMock(IConfig::class);
		$this->userManager = $this->createMock(IUserManager::class);
		$this->clientService = $this->createMock(IClientService::class);
		$this->client = $this->createMock(IClient::class);
		$this->groupManager = $this->createMock(IGroupManager::class);
		$this->logger = $this->createMock(LoggerInterface::class);

		$this->clientService->expects($this->once())->method('newClient')->willReturn($this->client);

		$expected = [
			1 => [
				'key' => 'paheko_url',
				'result' => 'paheko_url'
			],
			2 => [
				'key' => 'paheko_api_user',
				'result' => 'toto'
			],
			3 => [
				'key' => 'paheko_api_password',
				'result' => 'tata'
			],
			4 => [
				'key' => 'excluded_uids',
				'result' => 'framadmin'
			],
			5 => [
				'key' => 'admin_category_name',
				'default' => 'Administrateurs',
				'result' => 'Framadmins'
			],
		];

		$matcher = $this->exactly(5);
		$this->appConfig->expects($matcher)->method('getValueString')->willReturnCallback(function (string $appId, string $key) use ($expected, $matcher) {
			$callExpected = $expected[$matcher->numberOfInvocations()];
			$this->assertEquals(Application::APP_ID, $appId);
			$this->assertEquals($callExpected['key'], $key);
			return $callExpected['result'];
		});

		$this->syncUsers = new SyncUsers($this->appConfig, $this->config, $this->userManager, $this->clientService, $this->groupManager, $this->logger);
	}

	public const CREATE_USER_RESPONSE = <<<EOD
{
	"has_password": false,
	"has_otp": false,
	"has_pgp_key": false,
	"id": 5,
	"id_category": 1,
	"date_login": null,
	"date_updated": "2024-09-04 16:02:19",
	"id_parent": null,
	"is_parent": null,
	"preferences": null,
	"numero": 4,
	"nom": "coucou",
	"email": null,
	"adresse": null,
	"code_postal": null,
	"ville": null,
	"telephone": null,
	"lettre_infos": null,
	"date_inscription": null
}
EOD;


	#[DataProvider('dataTestCreateUser')]
	public function testCreateUser(bool $isAdmin): void {
		$user = $this->createMock(IUser::class);
		$user->expects($this->exactly(2))->method('getUID')->willReturn('uid');
		$user->expects($this->once())->method('getDisplayName')->willReturn('some user');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('user@example.com');
		$this->assertGetCategoriesCall();
		$this->groupManager->expects($this->once())->method('isAdmin')->willReturn($isAdmin);
		$response = $this->createMock(IResponse::class);
		$response->expects($this->once())->method('getBody')->willReturn(self::CREATE_USER_RESPONSE);
		$this->client->expects($this->once())->method('post')->with('paheko_url/api/user/new', [
			'body' => json_encode([
				'nom' => 'some user',
				'email' => 'user@example.com',
				'id_category' => $isAdmin ? 4 : 1
			]),
			'nextcloud' => [
				'allow_local_address' => true,
			],
			'headers' => [
				'Authorization' => 'Basic dG90bzp0YXRh',
				'Content-Type' => 'application/json'
			]
		])->willReturn($response);
		$this->config->expects($this->once())->method('setUserValue')->with('uid', Application::APP_ID, 'paheko_user_id', 5);
		$this->syncUsers->createUser($user);
	}

	public static function dataTestCreateUser(): array {
		return [
			[true],
			[false]
		];
	}

	public function testCreateUserWithExcludedUid() {
		$this->assertGetCategoriesCall();
		$user = $this->createMock(IUser::class);
		$user->expects($this->exactly(2))->method('getUID')->willReturn('framadmin');
		$this->groupManager->expects($this->never())->method('isAdmin');
		$this->syncUsers->createUser($user);
	}

	public function testCreateUserWithBadResponse() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->once())->method('getUID')->willReturn('uid');
		$user->expects($this->once())->method('getDisplayName')->willReturn('some user');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('user@example.com');
		$this->assertGetCategoriesCall();
		$this->groupManager->expects($this->once())->method('isAdmin')->willReturn(false);
		$this->client->expects($this->once())->method('post')->willThrowException(new \Exception('something bad happened', 500));
		$this->logger->expects($this->once())->method('error');
		$this->expectExceptionMessage('something bad happened');
		$this->syncUsers->createUser($user);
	}

	public function testCreateUserWithConflictResponse() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->exactly(2))->method('getUID')->willReturn('uid');
		$user->expects($this->once())->method('getDisplayName')->willReturn('some user');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('user@example.com');
		$this->assertGetCategoriesCall();
		$this->groupManager->expects($this->once())->method('isAdmin')->willReturn(false);
		$this->client->expects($this->once())->method('post')->willThrowException(new \Exception('something bad happened', 409));
		$matcher = $this->exactly(2);
		$this->logger->expects($matcher)->method('info')->willReturnCallback(function (string $message) use ($matcher) {
			match ($matcher->numberOfInvocations()) {
				1 => $this->assertEquals('Conflict while trying to create user uid on Paheko', $message),
				2 => $this->assertEquals('Created user uid on Paheko', $message)
			};
		});
		$this->syncUsers->createUser($user);
	}

	public function testUpdateUser() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->once())->method('getUID')->willReturn('uid');
		$user->expects($this->once())->method('getDisplayName')->willReturn('some user');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('user@example.com');
		$this->config->expects($this->once())->method('getUserValue')->with('uid', Application::APP_ID, 'paheko_user_id')->willReturn('6');
		$this->client->expects($this->once())->method('post')->with('paheko_url/api/user/6', [
			'body' => json_encode([
				'nom' => 'some user',
				'email' => 'user@example.com',
			]),
			'nextcloud' => [
				'allow_local_address' => true,
			],
			'headers' => [
				'Authorization' => 'Basic dG90bzp0YXRh',
				'Content-Type' => 'application/json'
			]
		]);
		$this->syncUsers->updateUser($user);
	}

	public function testUpdateUserWithHTTPError() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->once())->method('getUID')->willReturn('uid');
		$user->expects($this->once())->method('getDisplayName')->willReturn('some user');
		$user->expects($this->once())->method('getEMailAddress')->willReturn('user@example.com');
		$this->config->expects($this->once())->method('getUserValue')->with('uid', Application::APP_ID, 'paheko_user_id')->willReturn('6');
		$this->client->expects($this->once())->method('post')->willThrowException(new \Exception('something bad happened', 500));
		$this->logger->expects($this->once())->method('error');
		$this->expectExceptionMessage('something bad happened');
		$this->syncUsers->updateUser($user);
	}

	public function testDeleteUser() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->once())->method('getUID')->willReturn('uid');
		$this->config->expects($this->once())->method('getUserValue')->with('uid', Application::APP_ID, 'paheko_user_id')->willReturn('6');
		$this->client->expects($this->once())->method('delete')->with('paheko_url/api/user/6', [
			'nextcloud' => [
				'allow_local_address' => true,
			],
			'headers' => [
				'Authorization' => 'Basic dG90bzp0YXRh',
				'Content-Type' => 'application/json'
			]
		]);
		$this->syncUsers->deleteUser($user);
	}

	public function testDeleteUserWithBadResponse() {
		$user = $this->createMock(IUser::class);
		$user->expects($this->once())->method('getUID')->willReturn('uid');
		$this->config->expects($this->once())->method('getUserValue')->with('uid', Application::APP_ID, 'paheko_user_id')->willReturn('6');
		$this->client->expects($this->once())->method('delete')->willThrowException(new \Exception('something bad happened', 500));
		$this->logger->expects($this->once())->method('error');
		$this->expectExceptionMessage('something bad happened');
		$this->syncUsers->deleteUser($user);
	}

	public const CATEGORIES_RESPONSE = <<<EOD
{
	"3": {
		"id": 3,
		"name": "Administrateurs",
		"perm_web": 9,
		"perm_documents": 9,
		"perm_users": 9,
		"perm_accounting": 9,
		"perm_subscribe": 0,
		"perm_connect": 1,
		"perm_config": 9,
		"hidden": 0,
		"count": 0
	},
	"2": {
		"id": 2,
		"name": "Anciens membres",
		"perm_web": 0,
		"perm_documents": 0,
		"perm_users": 0,
		"perm_accounting": 0,
		"perm_subscribe": 0,
		"perm_connect": 0,
		"perm_config": 0,
		"hidden": 1,
		"count": 0
	},
	"4": {
		"id": 4,
		"name": "Framadmins",
		"perm_web": 0,
		"perm_documents": 0,
		"perm_users": 9,
		"perm_accounting": 9,
		"perm_subscribe": 0,
		"perm_connect": 1,
		"perm_config": 0,
		"hidden": 0,
		"count": 2
	},
	"1": {
		"id": 1,
		"name": "Membres actifs",
		"perm_web": 0,
		"perm_documents": 0,
		"perm_users": 0,
		"perm_accounting": 0,
		"perm_subscribe": 0,
		"perm_connect": 1,
		"perm_config": 0,
		"hidden": 0,
		"count": 2
	}
}
EOD;

	private function assertGetCategoriesCall(): void {
		$response = $this->createMock(IResponse::class);
		$response->expects($this->once())->method('getBody')->willReturn(self::CATEGORIES_RESPONSE);
		$this->client->expects($this->once())->method('get')->with('paheko_url/api/user/categories', [
			'nextcloud' => [
				'allow_local_address' => true,
			],
			'headers' => [
				'Authorization' => 'Basic dG90bzp0YXRh',
				'Content-Type' => 'application/json'
			]
		])->willReturn($response);
	}

	public function testInitialSyncAlreadyDone() {
		$this->appConfig->expects($this->once())->method('getValueBool')->with(Application::APP_ID, 'initial_user_import')->willReturn(true);
		$this->userManager->expects($this->never())->method('callForAllUsers');
		$this->syncUsers->initialSync();
	}

	public function testInitialSync() {
		$this->appConfig->expects($this->once())->method('getValueBool')->with(Application::APP_ID, 'initial_user_import')->willReturn(false);
		$this->assertGetCategoriesCall();
		$this->userManager->expects($this->once())->method('callForAllUsers');
		$this->appConfig->expects($this->once())->method('setValueBool')->with(Application::APP_ID, 'initial_user_import', true);
		$this->syncUsers->initialSync();
	}
}
